
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  bitstr.c 
  DESCR: 
  AUTH:  Xiaowan Huang
  DATE:  (C) Tue Nov 23 23:45:22 2004
*************************************************************************/

#include <string.h>
#include <assert.h>
#include "bitstr.h"

bitstr_t *
bitstr_alloc(int nbits) 
{
	bitstr_t *res = ALLOC(bitstr_t, 1);
	res->nbits = nbits;
	res->size = bitstr_size(res);
	res->ptr = ALLOC(byte_t, res->size);
	memset(res->ptr, 0, res->size);
	return res;
}

bitstr_t *
bitstr_dup(const bitstr_t *bstr)
{
	bitstr_t *res = bitstr_alloc(bitstr_n(bstr));
	memcpy(res->ptr, bstr->ptr, bstr->size);
	return res;
}

void
bitstr_free(bitstr_t *bstr)
{
	FREE(bstr->ptr);
	FREE(bstr);
}

void 
bitstr_expand(bitstr_t *bstr, int nbits)
{
	if (nbits < bstr->nbits)
		return;
	else if (bstr->size * CHAR_BIT >= nbits) {
		bstr->nbits = nbits;
		return;
	}
	else {
		byte_t *old_ptr = bstr->ptr;
		unsigned int old_size = bstr->size;
		bstr->size = (nbits-1) / CHAR_BIT + 1;
		bstr->nbits = nbits;
		bstr->ptr = ALLOC(byte_t, bstr->size);
		memcpy(bstr->ptr, old_ptr, old_size);
		memset(bstr->ptr+old_size, 0, bstr->size - old_size);
		FREE(old_ptr);
	}
}

void bitstr_append(bitstr_t *bstr, const bitstr_t *bstr2)
{
	unsigned int old_nbits = bstr->nbits;
	unsigned int old_size = (bstr->nbits-1)/CHAR_BIT + 1;
	byte_t *old_ptr = bstr->ptr;

	bstr->nbits += bstr2->nbits;
	bstr->size = bitstr_size(bstr);
	bstr->ptr = ALLOC(byte_t,  bstr->size);
	memcpy(bstr->ptr, old_ptr, old_size);
	FREE(old_ptr);
	
	unsigned int remainder = old_nbits % CHAR_BIT;
	if (remainder == 0) {
		memcpy(bstr->ptr + old_size, bstr2->ptr, (bstr2->nbits-1)/CHAR_BIT+1); 
	}
	else {
		byte_t mask = (0x1 << (CHAR_BIT - remainder)) - 1;
		bstr->ptr[old_size-1] &= ~mask;
		bstr->ptr[old_size-1] |= (bstr2->ptr[0] >> remainder) & mask;
		int i, from, to;
		for (i=old_size, from=CHAR_BIT-remainder; i<bstr->size; i++, from+=CHAR_BIT) {
			to = (from+CHAR_BIT > bstr2->nbits)? bstr2->nbits : from+CHAR_BIT;
			bstr->ptr[i] = (byte_t)bitstr_substr_to_uint32(bstr2, from, to);
		}
	}
}

bitstr_t* 
bitstr_concat(const bitstr_t *bstr1, const bitstr_t *bstr2)
{
	bitstr_t *res = bitstr_dup(bstr1);
	bitstr_append(res, bstr2);
	return res;
}

void 
bitstr_append_uint32(bitstr_t *bstr, unsigned long num)
{
	unsigned int old_nbits = bstr->nbits;
	unsigned int old_size = bitstr_size(bstr);

	bitstr_expand(bstr, bstr->nbits + 32);
	unsigned int remainder = old_nbits % CHAR_BIT;
	if (remainder == 0) {
		bstr->ptr[old_size] = (byte_t)((num >> 24) & 0xff);
		bstr->ptr[old_size+1] = (byte_t)((num >> 16) & 0xff); 
		bstr->ptr[old_size+2] = (byte_t)((num >> 8) & 0xff);
		bstr->ptr[old_size+3] = (byte_t)(num & 0xff);
		//memcpy(bstr->ptr + old_size, (byte_t *)&num, 4);
	}
	else {
		byte_t mask = (0x1 << (CHAR_BIT - remainder)) - 1;
		bstr->ptr[old_size-1] &= ~mask;
		bstr->ptr[old_size-1] |= ((byte_t)(num >> (24+remainder))) & mask;
		unsigned long tmp = num >> remainder;
		bstr->ptr[old_size] = (byte_t)((tmp & 0xff0000) >> 16);
		bstr->ptr[old_size+1] = (byte_t)((tmp & 0xff00) >> 8);
		bstr->ptr[old_size+2] = (byte_t)(tmp & 0xff);
		bstr->ptr[old_size+3] = (byte_t)((num & 0xff) << (CHAR_BIT - remainder));
	}
}

void 
bitstr_pack(bitstr_t *bstr)
{
	if (bstr->size > bitstr_size(bstr)) {
		bstr->size = bitstr_size(bstr);
		byte_t *old_ptr = bstr->ptr;
		bstr->ptr = ALLOC(byte_t, bstr->size);
		memcpy(bstr->ptr, old_ptr, bstr->size);
		FREE(old_ptr);
	}
}

void 
bitstr_abort(bitstr_t *bstr, int bit)
{
	assert(0);
}

bitstr_t *
bitstr_substr(const bitstr_t *bstr, int from, int to)
{
	assert(from <= to);
	assert(to <= bstr->nbits);
	
	bitstr_t *res = bitstr_alloc(to - from);
	int from_byte = from / CHAR_BIT;
	int to_byte = (to-1) / CHAR_BIT + 1;
	int shift = from % CHAR_BIT;

	int i, j;
	for (i = from_byte, j = 0; i < to_byte; i++) {
		byte_t l = bstr->ptr[i];
		byte_t r = (i<bstr->size)? bstr->ptr[i+1]:0;
		int tmp = (((int)l) << CHAR_BIT) + r;
		tmp <<= shift;
		res->ptr[j++] = (byte_t) ((tmp & 0xff00) >> CHAR_BIT);
	}

	return res;
}

bitstr_t *
bitstr_from_uint32(unsigned long num)
{
	bitstr_t *res = bitstr_alloc(sizeof(unsigned long) * CHAR_BIT);
	memcpy(res->ptr, &num, 4);
	return res;
}

bitstr_t *
bitstr_from_uint16(unsigned short num)
{
	bitstr_t *res = bitstr_alloc(sizeof(unsigned short) * CHAR_BIT);
	memcpy(res->ptr, &num, 2);
	return res;
}

// return the at most 32 bits from beginning.
unsigned long
bitstr_to_uint32(const bitstr_t *bstr)
{
	int to_byte = bitstr_size(bstr);
	if (to_byte > 4) 
		to_byte = 4;

	unsigned long res = 0L;
	int i;
	for (i=0; i<to_byte; ++i) {
		res <<= CHAR_BIT;
		res |= bstr->ptr[i];
	}

	return res;
}

unsigned long
bitstr_substr_to_uint32(const bitstr_t *bstr, int from, int to)
{
	assert(from <= to);
	assert(to <= bstr->nbits);
	assert(to - from <= 32);
	
	unsigned long res = 0L;
	int from_byte = from / CHAR_BIT;
	int to_byte = (to-1) / CHAR_BIT + 1;
	assert(to_byte - from_byte <= 4);
	int lshift = from % CHAR_BIT;
	int rshift = 32 - (to - from);
	
	int i, j;
	for (i=from_byte, j=32-CHAR_BIT; i<to_byte; i++, j-=CHAR_BIT) {
		byte_t l = bstr->ptr[i];
		byte_t r = (i<bstr->size)? bstr->ptr[i+1]:0;
		int tmp = ((((int)l) << CHAR_BIT) + r);
		tmp <<= lshift;
		tmp &= 0xff00;
		tmp >>= CHAR_BIT;
		//res <<= CHAR_BIT;
		//res |= (byte_t) ((tmp & 0xff00) >> CHAR_BIT);
		res |= (tmp << j);
	}
	
	return res >> rshift ;	
}

int
bitstr_bit_at(const bitstr_t *bstr, int at)
{
	assert(at < bstr->nbits);
	byte_t b = bstr->ptr[(at / CHAR_BIT)];
	byte_t t = 0x80 >> (at % CHAR_BIT);
	return (b & t)? 1:0;
}

/*
 *	Warning: not thread-reentriable!
 */
#define MAX_NBIT	4095

const char *
bitstr_to_bit_seq(const bitstr_t *bstr)
{
	static char sequence[MAX_NBIT + 1];
	int i;
	for (i=0; i<bstr->nbits && i<MAX_NBIT; ++i) {
		sequence[i] = (bitstr_bit_at(bstr, i) == 1)? '1': '0';
	}
	sequence[i] = '\0';
	return sequence;
}
