
/*========================================================================
               Copyright (C) 2004 by Xiaowan Huang
                            All rights reserved

    Permission is hereby granted, without written agreement and without
    license or royalty fees, to use, reproduce, prepare derivative
    works, distribute, and display this software and its documentation
    for any purpose, provided that (1) the above copyright notice and
    the following two paragraphs appear in all copies of the source code
    and (2) redistributions, including without limitation binaries,
    reproduce these notices in the supporting documentation. Substantial
    modifications to this software may be copyrighted by their authors
    and need not follow the licensing terms described here, provided
    that the new terms are clearly indicated in all files where they apply.
=======================================================================*/

/*************************************************************************
  FILE:  monitor.c 
  DESCR: 
  AUTH:  Xiaowan Huang
  DATE:  (C) Thu Dec 18 11:43:13 2004
*************************************************************************/

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <signal.h>
#include <assert.h>
#include <errno.h>

#include "concur.h"
#include "conlib.h"

static concur_manager_t *mgr = NULL;

static void concur_manager_sig_handler(int signo, siginfo_t *info, void *act)
{
	if (signo == SIGINT) {
		concur_print_manager_info(mgr);
		mgr->to_quit = 1;
	}
}

static void on_register_proc(const ctrl_msg_t *, ctrl_msg_t *);
static void on_unregister_proc(const ctrl_msg_t *, ctrl_msg_t *);
static void on_create_msg_queue(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg);
static void on_destroy_msg_queue(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg);
static void on_report_blocked(const ctrl_msg_t *recv_msg);
static void on_report_unblocked(const ctrl_msg_t *recv_msg);

int main(int argc, char *argv[])
{
	if (argc < 2) {
		printf("Usage: %s <socket_path>\n", argv[0]);
		return 0;
	}
	const char *socket_path = argv[1];
	mgr = concur_manager_init(socket_path);
	
	struct sigaction sigact;
	sigact.sa_sigaction = concur_manager_sig_handler;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags=SA_SIGINFO;
	DIE_ON_ERROR(sigaction(SIGINT, &sigact, NULL), "sigaction");
	
	while (!mgr->to_quit) {
		int sock;
		ctrl_msg_t recv_msg, send_msg;
		int count;
		struct sockaddr_un from_addr;
		int from_len;
	
		sock = mgr->lstn_sock;
		from_len = sizeof(struct sockaddr_un);
		count = recvfrom(sock, &recv_msg, sizeof(ctrl_msg_t), 0,
			(struct sockaddr *)&from_addr, &from_len);
		DIE_ON_ERROR(count, "recvfrom");
		printf("received %d bytes, from_len = %d\n", count, from_len);
		printf("received a message (%d) from process (%d)\n", 
			recv_msg.type, recv_msg.pid);
		
		send_msg.type = -1;
		switch (recv_msg.type) {
			case REGISTER_PROC: 
				on_register_proc(&recv_msg, &send_msg);
				break;
			case UNREGISTER_PROC:
				on_unregister_proc(&recv_msg, &send_msg);
				break;
			case CREATE_MSGQUEUE:
				on_create_msg_queue(&recv_msg, &send_msg);
				break;
			case DESTROY_MSGQUEUE:
				on_destroy_msg_queue(&recv_msg, &send_msg);
				break;
			case REPORT_BLOCKED:
				on_report_blocked(&recv_msg);
				break;
			case REPORT_UNBLOCKED:
				on_report_unblocked(&recv_msg);
				break;
			default:
				send_msg.type = NONE;
				send_msg.ctrl = NONE;
				send_msg.pid = recv_msg.pid;
				send_msg.val = 0;
				send_msg.error = EINVAL;
				printf("Unknown message type: %d\n", recv_msg.type);
		}
		
		if (send_msg.type > 0) {
			count = sendto(sock, &send_msg, sizeof(ctrl_msg_t), 0, 
				(struct sockaddr *)&from_addr, 
				sizeof(struct sockaddr_un));
			DIE_ON_ERROR(count, "sendto");
		}
		concur_print_manager_info(mgr);
	}

	concur_manager_finalize(mgr);
	unlink(socket_path);
	return 0;
}

static void on_register_proc(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg)
{
	int ret = concur_add_proc(mgr, recv_msg->pid);

	send_msg->type = recv_msg->type;
	send_msg->pid = recv_msg->pid;
	send_msg->val = 0;
	
	if (ret < 0) {
		send_msg->ctrl = REJECTED;
		send_msg->error = EEXIST;
	}
	else {
		send_msg->ctrl = GRANTED;
		send_msg->error = 0;
	}
}

static void on_unregister_proc(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg)
{
	int ret = concur_remove_proc(mgr, recv_msg->pid);

	send_msg->type = recv_msg->type;
	send_msg->pid = recv_msg->pid;
	send_msg->val = 0;

	if (ret < 0) {
		send_msg->ctrl = REJECTED;
		send_msg->error = ENOENT;
	}
	else {
		send_msg->ctrl = GRANTED;
		send_msg->error = 0;
	}
}

static void on_create_msg_queue(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg)
{
	int mlen = (recv_msg->val) >> 16;
	int msize = (recv_msg->val) & 0x00ff;
	int ret = concur_add_mqueue(mgr, mlen, msize, recv_msg->flag);

	send_msg->type = recv_msg->type;
	send_msg->pid = recv_msg->pid;
	if (ret < 0) {
		send_msg->val = -1;
		send_msg->ctrl = REJECTED;
		send_msg->error = -ret;
	}
	else {
		send_msg->val = ret;
		send_msg->ctrl = GRANTED;
		send_msg->error = 0;
	}
}

static void on_destroy_msg_queue(const ctrl_msg_t *recv_msg, ctrl_msg_t *send_msg)
{
	int ret = concur_remove_mqueue(mgr, recv_msg->val);

	send_msg->type = recv_msg->type;
	send_msg->pid = recv_msg->pid;
	send_msg->val = 0;
	
	if (ret < 0) {
		send_msg->ctrl = REJECTED;
		send_msg->error = -ret;
	}
	else {
		send_msg->ctrl = GRANTED;
		send_msg->error = 0;
	}
}

static void on_report_blocked(const ctrl_msg_t *recv_msg)
{
	concur_block_proc(mgr, recv_msg->pid);
}

static void on_report_unblocked(const ctrl_msg_t *recv_msg)
{
	concur_unblock_proc(mgr, recv_msg->pid);
}

