/* Needham-Schroeder public key */

/*
 * Only the following three steps of the protocol are modeled:
 *
 *   3. A->B: {NA,A}KB
 *   6. B->A: {NA,NB,B}KA       -- A assumes it is talking to B
 *   7. A->B: {NB}KB            -- B assumes it is talking to A
 *
 *   A: initiator, B: reponder
 */

/*
 * This version has the following characteristics:
 *  - Adversary always intercepts, honest agents only react to adversary.
 *  - Lowe's fix can be turned on and off.
 *  - Nonces allow multiple runs, but an honest agent may not participate 
 *      in more than one run concurrently.
 *  - There is only one adversary.
 *  - State variables of honest agents are global.  While untidy,
 *      this convention ensures that assertions of correctness can 
 *      be written.  (An alternative approach might use shared 
 *      variables for the honest agent state variables.)
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
/*****************************************************************************/
#define FIXED 0              /* turn Lowe's fix on and off */
#define QSZ 1                /* queue size for unidirectional channel */
#define NUMNONCES 2          /* number of nonces per agent, excluding the 
                                  undefined value 0 */
#define MAXKNOWLEDGE 6       /* number of messages remembered by adversary,
                                  must be at least 2 */
#define DUMP 0               /* allow adversary to erase from knowledge */

#define D 0                  /* D == dummy     */
#define I 1                  /* I == initiator */
#define R 2                  /* R == responder */
#define A 3                  /* A == adversary */
#define NUMAGENTS 4          /* for array indexing */

/*****************************************************************************/
typedef int AgentId;

typedef enum State_tag {SLEEP, WAIT} StateType;

typedef struct NonceType_tag {
  AgentId owner;
  int     value;
} NonceType;

typedef struct MsgType_tag {
  AgentId   key;           
  NonceType nonce1, nonce2;   
  AgentId   address;
} MsgType;

//MsgType * I_to_A;
int I_to_A_filled;
int I_to_A;
//MsgType * A_to_I;
int A_to_I_filled;
int A_to_I;
//MsgType * R_to_A;
int R_to_A_filled;
int R_to_A;
//MsgType * A_to_R;
int A_to_R_filled;
int A_to_R;

int        Ifresh; 
StateType  Istate; 
NonceType  Inonce; 
AgentId    Presponder; 
int        Rfresh; 
StateType  Rstate; 
NonceType  PInonce, Rnonce;  
AgentId    Pinitiator; 
int CommitHistory[NUMAGENTS][NUMNONCES + 1][NUMAGENTS][NUMNONCES + 1];

void mc2_call();
void Honest();
void InitiatorStep();
void ResponderStep();
void Adversary();
int VS_toss (int x);
void send_I_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void recv_I_to_A (MsgType * fmsg);
void send_A_to_I (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void recv_A_to_I (MsgType * fmsg);
void send_R_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void recv_R_to_A (MsgType * fmsg);
void send_A_to_R (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void recv_A_to_R (MsgType * fmsg);
void wait_process();
int VS_assert (int exp);

int main()
{
    key_t key_I_to_A;
    key_t key_A_to_I;
    key_t key_R_to_A;
    key_t key_A_to_R;
	
    Ifresh = 1;        
    Istate = SLEEP; 
    Inonce.owner = D; 
    Inonce.value = 0;  
    Presponder = D;    
    Rfresh = 1;        
    Rstate = SLEEP;
    PInonce.owner = D; 
    PInonce.value = 0;  
    Rnonce.owner = D; 
    Rnonce.value = 0;   
    Pinitiator = D;    
    I_to_A_filled = 0;
    A_to_I_filled = 0;
    R_to_A_filled = 0;
    A_to_R_filled = 0;
    if (fork() == 0) {
	    if ((key_I_to_A = ftok ("main.c", 'A')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_A_to_I = ftok ("main.c", 'B')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_R_to_A = ftok ("main.c", 'C')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_A_to_R = ftok ("main.c", 'D')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    
	    if ((I_to_A = msgget(key_I_to_A, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((A_to_I = msgget(key_A_to_I, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((R_to_A = msgget(key_R_to_A, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((A_to_R = msgget(key_A_to_R, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
 	    printf ("Adversary\n");
    	    Adversary();   
    } 
    else {
	    if ((key_I_to_A = ftok ("main.c", 'A')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_A_to_I = ftok ("main.c", 'B')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_R_to_A = ftok ("main.c", 'C')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    if ((key_A_to_R = ftok ("main.c", 'D')) == -1) {
		perror ("ftok");
		exit(1);
	    }
	    
	    if ((I_to_A = msgget(key_I_to_A, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((A_to_I = msgget(key_A_to_I, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((R_to_A = msgget(key_R_to_A, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
	    if ((A_to_R = msgget(key_A_to_R, 0644|IPC_CREAT)) == -1) {
		perror("msgget");
		exit(1);
	    }
      	    Honest();      
    }
}

void Honest()
{
  int rand; 
  printf ("honest\n");
  while (1) {
    rand = VS_toss (1);
    if (rand == 0) {
  	printf ("Initiator\n");
    	InitiatorStep();
    } else {
 	 printf ("Responder\n");
  	ResponderStep();
    }
  }
}

void InitiatorStep()
{
  MsgType  msg[1];
  int rand;

  if (Istate == SLEEP && Ifresh <= NUMNONCES) {
    rand = VS_toss (1);
    if (rand == 0) {
    	Presponder = R;
    } else {
	Presponder = A;
    }
    Inonce.owner = I;
    Inonce.value = Ifresh;
    ++Ifresh;
    {
	AgentId a1, a2, a4, a6;
	int a3, a5;

	a1 = Presponder;
	a2 = Inonce.owner;
	a3 = Inonce.value;
	a4 = D;
	a5 = 0;
	a6 = I;
	send_I_to_A (a1, a2, a3, a4, a5, a6);
	I_to_A_filled = 1;
    }
    Istate = WAIT;
  } else if (Istate == WAIT) {
	{
		recv_A_to_I (msg);
	        A_to_I_filled = 0;
        }
      	if (msg[0].key == I 
      		&& msg[0].nonce1.owner == Inonce.owner && msg[0].nonce1.value == Inonce.value
      		&& !(msg[0].nonce2.owner == D 
        	|| msg[0].nonce2.value <= 0
           	|| msg[0].nonce2.value > NUMNONCES) 
      		&& (!FIXED || (msg[0].address != D))     
      		&& (FIXED || (msg[0].address == D))) {

	      	if (Presponder == R) {
			int exp;
			exp = Rstate == WAIT && Pinitiator == I
		  			&& Rnonce.owner == msg[0].nonce2.owner && Rnonce.value == msg[0].nonce2.value 
		  			&& PInonce.owner == Inonce.owner && PInonce.value == Inonce.value;
			assert (exp);
	        }

	      	CommitHistory[Inonce.owner][Inonce.value][Presponder][Rnonce.value] = 1;      
	      	msg[0].key = Presponder;
	      	msg[0].nonce1.owner = msg[0].nonce2.owner;
	     	msg[0].nonce1.value = msg[0].nonce2.value;
	     	msg[0].nonce2.owner = D;
	     	msg[0].nonce2.value = 0;
	      	msg[0].address = D;
		{
			AgentId a1, a2, a4, a6;
			int a3, a5;

			a1 = msg[0].key;
			a2 = msg[0].nonce1.owner;
			a3 = msg[0].nonce1.value;
			a4 = msg[0].nonce2.owner;
			a5 = msg[0].nonce2.value;
			a6 = msg[0].address;
			send_I_to_A (a1, a2, a3, a4, a5, a6);
			I_to_A_filled = 1;
	     	} 
	      	Istate = SLEEP;
    	}
  }
}

void ResponderStep()
{
  MsgType  msg[1];

  if (Rstate == SLEEP && Rfresh <= NUMNONCES ) {
      {
	      recv_A_to_R (msg);
	      A_to_R_filled = 0;
      }
      if (msg[0].key == R && !(msg[0].nonce1.owner == D 
           || msg[0].nonce1.value <= 0
           || msg[0].nonce1.value > NUMNONCES) 
      	   && (msg[0].nonce2.owner == D 
           || msg[0].nonce2.value <= 0
           || msg[0].nonce2.value > NUMNONCES) 
     	   && msg[0].address != D
      	   && msg[0].address != R) {
	      Rnonce.owner = R;
	      Rnonce.value = Rfresh;
	      ++Rfresh;
	      PInonce.owner = msg[0].nonce1.owner;
	      PInonce.value = msg[0].nonce1.value;
	      Pinitiator = msg[0].address;
	   
	      msg[0].key = Pinitiator;
	      msg[0].nonce1.owner = PInonce.owner;
	      msg[0].nonce1.value = PInonce.value;
	      msg[0].nonce2.owner = Rnonce.owner;
	      msg[0].nonce2.value = Rnonce.value;
	      if (FIXED) {
		msg[0].address = R;
	      }
	      else {
		msg[0].address = D;
	      }
    	     {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_R_to_A (a1, a2, a3, a4, a5, a6);
		R_to_A_filled = 1;
	     }
	      Rstate = WAIT;
    	}
  } else if (Rstate == WAIT /* && !is_queue_empty(A_to_R, QSZ) */) {
      {
	      recv_A_to_R (msg);
	      A_to_R_filled = 0;
      } 
      if (msg[0].key == R 
      	  && msg[0].nonce1.owner == Rnonce.owner && msg[0].nonce1.value == Rnonce.value
      	  && (msg[0].nonce2.owner == D 
          || msg[0].nonce2.value <= 0
          || msg[0].nonce2.value > NUMNONCES) 
     	  && msg[0].address == D) {
		
	      if (Pinitiator == I) {
		int exp;
		exp = CommitHistory[Pinitiator][PInonce.value][Rnonce.owner][Rnonce.value] == 1;
		assert (exp);
	      }

      	      Rstate = SLEEP;
    	}
  }
}

void Adversary()
{
  int       nonces[NUMAGENTS][NUMNONCES + 1];   /* known nonces   */
  MsgType   messages[MAXKNOWLEDGE];                   /* known messages */ 
  int       nextspace;          /* next available space in messages */
  MsgType msg[1];
  int rand;

  while (1) {
    rand = VS_toss (2);
    if (rand == 0) {
        int successful = 0;

        if (nextspace < MAXKNOWLEDGE) {
          if (VS_toss(1) == 0) {
	      {
	      	      recv_I_to_A (msg);
		      I_to_A_filled = 0;
	      }
              successful = 1;
	  } else {
              {
	      	      recv_R_to_A (msg);
		      R_to_A_filled = 0;
	      }
              successful = 1;
	  }
          if (successful) {
              messages[nextspace].key = msg[0].key;
              messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
              messages[nextspace].nonce1.value = msg[0].nonce1.value;
              messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
              messages[nextspace].nonce2.value = msg[0].nonce2.value;
	      messages[nextspace].address = msg[0].address;
              ++nextspace;
              if (msg[0].key == A) {
              	nonces[msg[0].nonce1.owner][msg[0].nonce1.value] = 1;
             	nonces[msg[0].nonce2.owner][msg[0].nonce2.value] = 1;             
	      }
	  }
	}
     } else if (rand == 1) {
        int  target = VS_toss(MAXKNOWLEDGE - 1);

        if (target < nextspace) {
	      msg[0].key = messages[target].key;
	      msg[0].nonce1.owner = messages[target].nonce1.owner;
	      msg[0].nonce1.value = messages[target].nonce1.value;
	      msg[0].nonce2.owner = messages[target].nonce2.owner;
	      msg[0].nonce2.value = messages[target].nonce2.value;
	      msg[0].address = messages[target].address;
          if (VS_toss(1) == 0) {
	   {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_A_to_I (a1, a2, a3, a4, a5, a6);
		A_to_I_filled = 1;
	   }
	} else {
	   {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_A_to_R (a1, a2, a3, a4, a5, a6);
		A_to_R_filled = 1;
	   }
	}
      }
    } else if (rand == 2) {
        if (nextspace >= MAXKNOWLEDGE && DUMP) {  
          messages[VS_toss(MAXKNOWLEDGE - 2)].key = messages[MAXKNOWLEDGE - 1].key;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce1.owner = messages[MAXKNOWLEDGE - 1].nonce1.owner;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce1.value = messages[MAXKNOWLEDGE - 1].nonce1.value;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce2.owner = messages[MAXKNOWLEDGE - 1].nonce2.owner;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce2.value = messages[MAXKNOWLEDGE - 1].nonce2.value;
          messages[VS_toss(MAXKNOWLEDGE - 2)].address = messages[MAXKNOWLEDGE - 1].address;
          nextspace = MAXKNOWLEDGE - 1;
	}
        else if (nextspace < MAXKNOWLEDGE) {
          int     duplicate = 0, i;
          AgentId k, n1o, n2o, a;
          int     n1v, n2v;

          k = VS_toss(NUMAGENTS - 3) + 1; 
          if (k == I) {
            n1o = I;
            n1v = VS_toss(NUMNONCES - 1) + 1; /* n1v in 1..NUMNONCES */
            if (VS_toss(1) == 0) {
            	n2o = R;
	    } else {
            	n2o = A;
	    }
            n2v = (VS_toss(NUMNONCES - 1)) + 1;
	    if (!FIXED) {
              a = D;
	    }
            else {
              if (VS_toss(1) == 0) {
              	a = R;
	      } else {
              	a = A;
	      }
	    }
	    if (nonces[n1o][n1v] && nonces[n2o][n2v]) {  /* nonces are known */
		    msg[0].key = k;
		    msg[0].nonce1.owner = n1o;
		    msg[0].nonce1.value = n1v;
		    msg[0].nonce2.owner = n2o;
		    msg[0].nonce2.value = n2v;
		    msg[0].address = a;
		    for (i = 0; i < nextspace; ++i) {
		      	duplicate = duplicate || (msg[0].key == messages[i].key 
						  && msg[0].nonce1.owner == messages[i].nonce1.owner
						  && msg[0].nonce1.value == messages[i].nonce1.value
						  && msg[0].nonce2.owner == messages[i].nonce2.owner
						  && msg[0].nonce1.value == messages[i].nonce2.value);
		
		    }
		    if (!duplicate) {
		    	messages[nextspace].key = msg[0].key;
		    	messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
		    	messages[nextspace].nonce1.value = msg[0].nonce1.value;
		    	messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
		    	messages[nextspace].nonce2.value = msg[0].nonce2.value;
		    	messages[nextspace].address = msg[0].address;
		    	++nextspace;
		    }
	    }
	  }
          else if (k == R) {
            n2o = D;
            n2v = 0;

            if (VS_toss(1) == 0) {
              a = D;
              n1o = R;
              n1v = VS_toss(NUMNONCES - 1) + 1;
	    } else {
              if (VS_toss(1) == 0) {
              	a = I;
	       } else {
              	a = A;
	       }
              if (VS_toss(1) == 0) {
              	n1o = I;
	      }else {
              	n1o = A;
              }
              n1v = VS_toss(NUMNONCES - 1) + 1;                  
            }
	    if (nonces[n1o][n1v] && nonces[n2o][n2v])  {  /* nonces are known */
	      msg[0].key = k;
	      msg[0].nonce1.owner = n1o;
	      msg[0].nonce1.value = n1v;
	      msg[0].nonce2.owner = n2o;
	      msg[0].nonce2.value = n2v;
	      msg[0].address = a;
	      for (i = 0; i < nextspace; ++i) {
	      		duplicate = duplicate || (msg[0].key == messages[i].key 
					  && msg[0].nonce1.owner == messages[i].nonce1.owner
					  && msg[0].nonce1.value == messages[i].nonce1.value
					  && msg[0].nonce2.owner == messages[i].nonce2.owner
					  && msg[0].nonce1.value == messages[i].nonce2.value);
	      }
	      if (!duplicate) {
		    messages[nextspace].key = msg[0].key;
		    messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
		    messages[nextspace].nonce1.value = msg[0].nonce1.value;
		    messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
		    messages[nextspace].nonce2.value = msg[0].nonce2.value;
		    messages[nextspace].address = msg[0].address;
		    ++nextspace;
	       }
	     }
          }
        }
    }
  }
}

int VS_toss (int x) {
	int ret;
	ret = ((x+1) * (rand()/(RAND_MAX + 1.0)));
	return ret;

}

void send_I_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
        MsgType  msg;
	ssize_t ret_val;
	msg.key = a1;
	msg.nonce1.owner = a2;
	msg.nonce1.value = a3;
	msg.nonce2.owner = a4;
	msg.nonce2.value = a5;
	msg.address = a6;
	printf ("msgsnd I_to_A\n");
	ret_val = msgsnd (I_to_A, (MsgType*)&msg, sizeof (msg), IPC_NOWAIT); 
	printf ("%d\n", ret_val);
	while (ret_val == EAGAIN) {
		mc2_call();
		ret_val = msgsnd (I_to_A, (MsgType*)&msg, sizeof (msg), IPC_NOWAIT); 
	}
}
void recv_I_to_A (MsgType * fmsg) {
        MsgType msg;
	while (msgrcv (I_to_A, (MsgType*)&msg, sizeof (msg), 0, IPC_NOWAIT) == EAGAIN) {
		perror("msgrcv");
		mc2_call();
	}
	fmsg->key = msg.key;
	fmsg->nonce1.owner = msg.nonce1.owner;
	fmsg->nonce1.value = msg.nonce1.value;
	fmsg->nonce2.owner = msg.nonce2.owner;
	fmsg->nonce2.value = msg.nonce2.value;
	fmsg->address = msg.address;
}

void send_A_to_I (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
        MsgType  msg;
	msg.key = a1;
	msg.nonce1.owner = a2;
	msg.nonce1.value = a3;
	msg.nonce2.owner = a4;
	msg.nonce2.value = a5;
	msg.address = a6;
	while (msgsnd (A_to_I, (MsgType*)&msg, sizeof (msg), IPC_NOWAIT) == EAGAIN) {
		perror("msgsnd");
		mc2_call();
	}
}
void recv_A_to_I (MsgType * fmsg) {
        MsgType msg;
	while (msgrcv (A_to_I, (MsgType*)&msg, sizeof (msg), 0, IPC_NOWAIT) == EAGAIN) {
		perror("msgrcv");
		mc2_call();
	}
	fmsg->key = msg.key;
	fmsg->nonce1.owner = msg.nonce1.owner;
	fmsg->nonce1.value = msg.nonce1.value;
	fmsg->nonce2.owner = msg.nonce2.owner;
	fmsg->nonce2.value = msg.nonce2.value;
	fmsg->address = msg.address;
}

void send_R_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
        MsgType  msg;
	msg.key = a1;
	msg.nonce1.owner = a2;
	msg.nonce1.value = a3;
	msg.nonce2.owner = a4;
	msg.nonce2.value = a5;
	msg.address = a6;
	while (msgsnd (R_to_A, (MsgType*)&msg, sizeof (msg), IPC_NOWAIT) == EAGAIN) {
		perror("msgsnd");
		mc2_call();
	}
}
void recv_R_to_A (MsgType * fmsg) {
        MsgType msg;
	while (msgrcv (R_to_A, (MsgType*)&msg, sizeof (msg), 0, IPC_NOWAIT) == EAGAIN) {
		perror("msgrcv");
		mc2_call();
	}
	fmsg->key = msg.key;
	fmsg->nonce1.owner = msg.nonce1.owner;
	fmsg->nonce1.value = msg.nonce1.value;
	fmsg->nonce2.owner = msg.nonce2.owner;
	fmsg->nonce2.value = msg.nonce2.value;
	fmsg->address = msg.address;
}

void send_A_to_R (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
        MsgType  msg;
	msg.key = a1;
	msg.nonce1.owner = a2;
	msg.nonce1.value = a3;
	msg.nonce2.owner = a4;
	msg.nonce2.value = a5;
	msg.address = a6;
	while (msgsnd (A_to_R, (MsgType*)&msg, sizeof (msg), IPC_NOWAIT) == EAGAIN) {
		perror("msgsnd");
		mc2_call();
	}
}
void recv_A_to_R (MsgType * fmsg) {
        MsgType msg;
	ssize_t ret_val;
	ret_val = msgrcv (A_to_R, (MsgType*)&msg, sizeof (msg), 0, IPC_NOWAIT);
	printf("errno = %d\n", errno);
	while (ret_val < 0 && errno == ENOMSG) {
		mc2_call();
		ret_val = msgrcv (A_to_R, (MsgType*)&msg, sizeof (msg), 0, IPC_NOWAIT);
	}
	fmsg->key = msg.key;
	fmsg->nonce1.owner = msg.nonce1.owner;
	fmsg->nonce1.value = msg.nonce1.value;
	fmsg->nonce2.owner = msg.nonce2.owner;
	fmsg->nonce2.value = msg.nonce2.value;
	fmsg->address = msg.address;
}

void mc2_call(){
	printf ("exiting in mc2_call\n");
	exit(1);	
}
void wait_process() {

}

int VS_assert (int exp) {
	if (exp == 0)
		return 0;
	else 
		return 1;
}
