/* Needham-Schroeder public key */

/*
 * Only the following three steps of the protocol are modeled:
 *
 *   3. A->B: {NA,A}KB
 *   6. B->A: {NA,NB,B}KA       -- A assumes it is talking to B
 *   7. A->B: {NB}KB            -- B assumes it is talking to A
 *
 *   A: initiator, B: reponder
 */

/*
 * This version has the following characteristics:
 *  - Adversary always intercepts, honest agents only react to adversary.
 *  - Lowe's fix can be turned on and off.
 *  - Nonces allow multiple runs, but an honest agent may not participate 
 *      in more than one run concurrently.
 *  - There is only one adversary.
 *  - State variables of honest agents are global.  While untidy,
 *      this convention ensures that assertions of correctness can 
 *      be written.  (An alternative approach might use shared 
 *      variables for the honest agent state variables.)
 */

/*****************************************************************************/
#define FIXED 0              /* turn Lowe's fix on and off */
#define QSZ 1                /* queue size for unidirectional channel */
#define NUMNONCES 2          /* number of nonces per agent, excluding the 
                                  undefined value 0 */
#define MAXKNOWLEDGE 6       /* number of messages remembered by adversary,
                                  must be at least 2 */
#define DUMP 0               /* allow adversary to erase from knowledge */

#define D 0                  /* D == dummy     */
#define I 1                  /* I == initiator */
#define R 2                  /* R == responder */
#define A 3                  /* A == adversary */
#define NUMAGENTS 4          /* for array indexing */

/*****************************************************************************/
typedef int AgentId;

typedef enum State_tag {SLEEP, WAIT} StateType;

typedef struct NonceType_tag {
  AgentId owner;
  int     value;
} NonceType;

typedef struct MsgType_tag {
  AgentId   key;              /* encryption key */
  NonceType nonce1, nonce2;   
  AgentId   address;
} MsgType;

MsgType shared_channel_I_to_A[1];
int shared_channel_I_to_A_filled;
MsgType shared_channel_A_to_I[1];
int shared_channel_A_to_I_filled;
MsgType shared_channel_R_to_A[1];
int shared_channel_R_to_A_filled;
MsgType shared_channel_A_to_R[1];
int shared_channel_A_to_R_filled;

int        Ifresh; 
StateType  Istate; 
NonceType  Inonce; 
AgentId    Presponder; 
int        Rfresh; 
StateType  Rstate; 
NonceType  PInonce, Rnonce;  
AgentId    Pinitiator; 
int CommitHistory[NUMAGENTS][NUMNONCES + 1][NUMAGENTS][NUMNONCES + 1];

void Honest();
void InitiatorStep();
void ResponderStep();
void Adversary();
int VS_toss (int x);
void send_shared_channel_I_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void send_shared_channel_A_to_I (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void send_shared_channel_R_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void send_shared_channel_A_to_R (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6);
void wait_process();
int VS_assert (int exp);

int main()
{
    Ifresh = 1;        
    Istate = SLEEP; 
    Inonce.owner = D; 
    Inonce.value = 0;  
    Presponder = D;    
    Rfresh = 1;        
    Rstate = SLEEP;
    PInonce.owner = D; 
    PInonce.value = 0;  
    Rnonce.owner = D; 
    Rnonce.value = 0;   
    Pinitiator = D;    
    shared_channel_I_to_A_filled = 0;
    shared_channel_A_to_I_filled = 0;
    shared_channel_R_to_A_filled = 0;
    shared_channel_A_to_R_filled = 0;
   shared_channel_A_to_R[0].nonce2.value = 1;
    if (fork() == 0) {
    	Adversary();   /* process 2 */
    } 
    else {
      	Honest();      /* process 1 */
    }
}

void Honest()
{
  int rand;
  while (1) {
    rand = VS_toss (1);
    if (rand == 0) {
    	InitiatorStep();
    } else {
  	ResponderStep();
    }
  }
}

void InitiatorStep()
{
  MsgType  msg[1];
  int rand;

  if (Istate == SLEEP && Ifresh <= NUMNONCES) {
    /* start protocol */
    rand = VS_toss (1);
    if (rand == 0) {
    	Presponder = R;
    } else {
	Presponder = A;
    }
    Inonce.owner = I;
    Inonce.value = Ifresh;
    ++Ifresh;
    /*if (shared_channel_I_to_A_filled == 0) {
	AgentId a1, a2, a4, a6;
	int a3, a5;

	a1 = Presponder;
	a2 = Inonce.owner;
	a3 = Inonce.value;
	a4 = D;
	a5 = 0;
	a6 = I;
	send_shared_channel_I_to_A (a1, a2, a3, a4, a5, a6);
	shared_channel_I_to_A_filled = 1;
	//wait_process();
    } else {
	wait_process();
    }
   */
    while (shared_channel_I_to_A_filled != 0) {
	wait_process();
    }
    {
	AgentId a1, a2, a4, a6;
	int a3, a5;

	a1 = Presponder;
	a2 = Inonce.owner;
	a3 = Inonce.value;
	a4 = D;
	a5 = 0;
	a6 = I;
	send_shared_channel_I_to_A (a1, a2, a3, a4, a5, a6);
	shared_channel_I_to_A_filled = 1;
	wait_process();
    }
    Istate = WAIT;
  } else if (Istate == WAIT) {
	/*
	if(shared_channel_A_to_I_filled == 1) {
		msg[0].key = shared_channel_A_to_I[0].key;
		msg[0].nonce1.owner = shared_channel_A_to_I[0].nonce1.owner;
		msg[0].nonce1.value = shared_channel_A_to_I[0].nonce1.value;
		msg[0].nonce2.owner = shared_channel_A_to_I[0].nonce2.owner;
		msg[0].nonce2.value = shared_channel_A_to_I[0].nonce2.value;
		msg[0].address = shared_channel_A_to_I[0].address;
		shared_channel_A_to_I_filled = 0;
        } else {
	      wait_process();
        }
	*/
	while (shared_channel_A_to_I_filled != 1) {
		wait_process();
	}
	{
		msg[0].key = shared_channel_A_to_I[0].key;
		msg[0].nonce1.owner = shared_channel_A_to_I[0].nonce1.owner;
		msg[0].nonce1.value = shared_channel_A_to_I[0].nonce1.value;
		msg[0].nonce2.owner = shared_channel_A_to_I[0].nonce2.owner;
		msg[0].nonce2.value = shared_channel_A_to_I[0].nonce2.value;
		msg[0].address = shared_channel_A_to_I[0].address;
		shared_channel_A_to_I_filled = 0;
        }
      	if (msg[0].key == I 
      		&& msg[0].nonce1.owner == Inonce.owner && msg[0].nonce1.value == Inonce.value
      		&& !(msg[0].nonce2.owner == D 
        	|| msg[0].nonce2.value <= 0
           	|| msg[0].nonce2.value > NUMNONCES) /* defined */
      		&& (!FIXED || (msg[0].address != D))     
      		&& (FIXED || (msg[0].address == D))) {

	      	if (Presponder == R) {
			int exp;
			exp = Rstate == WAIT && Pinitiator == I
		  			&& Rnonce.owner == msg[0].nonce2.owner && Rnonce.value == msg[0].nonce2.value 
		  			&& PInonce.owner == Inonce.owner && PInonce.value == Inonce.value;
			VS_assert (exp);
	        }

	      	CommitHistory[Inonce.owner][Inonce.value][Presponder][Rnonce.value] = 1;      
	      	msg[0].key = Presponder;
	      	msg[0].nonce1.owner = msg[0].nonce2.owner;
	     	msg[0].nonce1.value = msg[0].nonce2.value;
	     	msg[0].nonce2.owner = D;
	     	msg[0].nonce2.value = 0;
	      	msg[0].address = D;
		/*
	     	if (shared_channel_I_to_A_filled == 0) {
			AgentId a1, a2, a4, a6;
			int a3, a5;

			a1 = msg[0].key;
			a2 = msg[0].nonce1.owner;
			a3 = msg[0].nonce1.value;
			a4 = msg[0].nonce2.owner;
			a5 = msg[0].nonce2.value;
			a6 = msg[0].address;
			send_shared_channel_I_to_A (a1, a2, a3, a4, a5, a6);
			shared_channel_I_to_A_filled = 1;
			//wait_process();
	     	} else {
			wait_process();
	     	}
		*/
	     	while (shared_channel_I_to_A_filled != 0) {
			wait_process ();
		}
		{
			AgentId a1, a2, a4, a6;
			int a3, a5;

			a1 = msg[0].key;
			a2 = msg[0].nonce1.owner;
			a3 = msg[0].nonce1.value;
			a4 = msg[0].nonce2.owner;
			a5 = msg[0].nonce2.value;
			a6 = msg[0].address;
			send_shared_channel_I_to_A (a1, a2, a3, a4, a5, a6);
			shared_channel_I_to_A_filled = 1;
			wait_process();
	     	} 
	      	Istate = SLEEP;
    	}
  }
}

void ResponderStep()
{
  MsgType  msg[1];

  if (Rstate == SLEEP && Rfresh <= NUMNONCES ) {
      /*
      if (shared_channel_A_to_R_filled == 1) {
	      msg[0].key = shared_channel_A_to_R[0].key;
	      msg[0].nonce1.owner = shared_channel_A_to_R[0].nonce1.owner;
	      msg[0].nonce1.value = shared_channel_A_to_R[0].nonce1.value;
	      msg[0].nonce2.owner = shared_channel_A_to_R[0].nonce2.owner;
	      msg[0].nonce2.value = shared_channel_A_to_R[0].nonce2.value;
	      msg[0].address = shared_channel_A_to_R[0].address;
	      shared_channel_A_to_R_filled = 0;
      } else {
	      wait_process();
      }
      */
      while (shared_channel_A_to_R_filled != 1) {
	wait_process();
      }
      {
	      msg[0].key = shared_channel_A_to_R[0].key;
	      msg[0].nonce1.owner = shared_channel_A_to_R[0].nonce1.owner;
	      msg[0].nonce1.value = shared_channel_A_to_R[0].nonce1.value;
	      msg[0].nonce2.owner = shared_channel_A_to_R[0].nonce2.owner;
	      msg[0].nonce2.value = shared_channel_A_to_R[0].nonce2.value;
	      msg[0].address = shared_channel_A_to_R[0].address;
	      shared_channel_A_to_R_filled = 0;
      }
      if (msg[0].key == R && !(msg[0].nonce1.owner == D 
           || msg[0].nonce1.value <= 0
           || msg[0].nonce1.value > NUMNONCES) /* defined */
      	   && (msg[0].nonce2.owner == D 
           || msg[0].nonce2.value <= 0
           || msg[0].nonce2.value > NUMNONCES) /* undefined */
     	   && msg[0].address != D
      	   && msg[0].address != R) {
      /* msg has correct form; begin protocol */
	      Rnonce.owner = R;
	      Rnonce.value = Rfresh;
	      ++Rfresh;
	      PInonce.owner = msg[0].nonce1.owner;
	      PInonce.value = msg[0].nonce1.value;
	      Pinitiator = msg[0].address;
	   
	      msg[0].key = Pinitiator;
	      msg[0].nonce1.owner = PInonce.owner;
	      msg[0].nonce1.value = PInonce.value;
	      msg[0].nonce2.owner = Rnonce.owner;
	      msg[0].nonce2.value = Rnonce.value;
	      if (FIXED) {
		msg[0].address = R;
	      }
	      else {
		msg[0].address = D;
	      }
	    
	     /*
	     if (shared_channel_R_to_A_filled == 0) {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_R_to_A (a1, a2, a3, a4, a5, a6);
		shared_channel_R_to_A_filled = 1;
		//wait_process();
	     } else {
		wait_process();
	     }
	     */
	     while (shared_channel_R_to_A_filled != 0) {
		wait_process ();
	     }
    	     {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_R_to_A (a1, a2, a3, a4, a5, a6);
		shared_channel_R_to_A_filled = 1;
		wait_process();
	     }
	      Rstate = WAIT;
    	}
  } else if (Rstate == WAIT /* && !is_queue_empty(shared_channel_A_to_R, QSZ) */) {
      /*
      if (shared_channel_A_to_R_filled == 1) {
	      msg[0].key = shared_channel_A_to_R[0].key;
	      msg[0].nonce1.owner = shared_channel_A_to_R[0].nonce1.owner;
	      msg[0].nonce1.value = shared_channel_A_to_R[0].nonce1.value;
	      msg[0].nonce2.owner = shared_channel_A_to_R[0].nonce2.owner;
	      msg[0].nonce2.value = shared_channel_A_to_R[0].nonce2.value;
	      msg[0].address = shared_channel_A_to_R[0].address;
	      shared_channel_A_to_R_filled = 0;
      } else {
	      wait_process();
      } 
      */	
      while (shared_channel_A_to_R_filled != 1) {
	wait_process();
      } 
      {
	      msg[0].key = shared_channel_A_to_R[0].key;
	      msg[0].nonce1.owner = shared_channel_A_to_R[0].nonce1.owner;
	      msg[0].nonce1.value = shared_channel_A_to_R[0].nonce1.value;
	      msg[0].nonce2.owner = shared_channel_A_to_R[0].nonce2.owner;
	      msg[0].nonce2.value = shared_channel_A_to_R[0].nonce2.value;
	      msg[0].address = shared_channel_A_to_R[0].address;
	      shared_channel_A_to_R_filled = 0;
      } 
      if (msg[0].key == R 
      	  && msg[0].nonce1.owner == Rnonce.owner && msg[0].nonce1.value == Rnonce.value
      	  && (msg[0].nonce2.owner == D 
          || msg[0].nonce2.value <= 0
          || msg[0].nonce2.value > NUMNONCES) /* undefined */
     	  && msg[0].address == D) {
		
	      if (Pinitiator == I) {
		int exp;
		exp = CommitHistory[Pinitiator][PInonce.value][Rnonce.owner][Rnonce.value] == 1;
		VS_assert (exp);
	      }

      	      Rstate = SLEEP;
    	}
  }
}

void Adversary()
{
  int       nonces[NUMAGENTS][NUMNONCES + 1];   /* known nonces   */
  MsgType   messages[MAXKNOWLEDGE];                   /* known messages */ 
  int       nextspace;          /* next available space in messages */
  MsgType msg[1];
  int rand;

  /* can give adversary some initial nonces here */
  while (1) {
    rand = VS_toss (2);
    if (rand == 0) {
     /* attempt to receive msg */
        int successful = 0;

        if (nextspace < MAXKNOWLEDGE) {
          if (VS_toss(1) == 0) {
	      /*
              if(shared_channel_I_to_A_filled == 1) {
		      msg[0].key = shared_channel_I_to_A[0].key;
		      msg[0].nonce1.owner = shared_channel_I_to_A[0].nonce1.owner;
		      msg[0].nonce1.value = shared_channel_I_to_A[0].nonce1.value;
		      msg[0].nonce2.owner = shared_channel_I_to_A[0].nonce2.owner;
		      msg[0].nonce2.value = shared_channel_I_to_A[0].nonce2.value;
		      msg[0].address = shared_channel_I_to_A[0].address;
		      shared_channel_I_to_A_filled = 0;
	      } else {
		      wait_process();
	      }
	      */
              while (shared_channel_I_to_A_filled != 1) {
		wait_process();
	      }
	      {
		      msg[0].key = shared_channel_I_to_A[0].key;
		      msg[0].nonce1.owner = shared_channel_I_to_A[0].nonce1.owner;
		      msg[0].nonce1.value = shared_channel_I_to_A[0].nonce1.value;
		      msg[0].nonce2.owner = shared_channel_I_to_A[0].nonce2.owner;
		      msg[0].nonce2.value = shared_channel_I_to_A[0].nonce2.value;
		      msg[0].address = shared_channel_I_to_A[0].address;
		      shared_channel_I_to_A_filled = 0;
	      }
              successful = 1;
	  } else {
	      /*
	      if (shared_channel_R_to_A_filled == 1) { 
		      msg[0].key = shared_channel_R_to_A[0].key;
		      msg[0].nonce1.owner = shared_channel_R_to_A[0].nonce1.owner;
		      msg[0].nonce1.value = shared_channel_R_to_A[0].nonce1.value;
		      msg[0].nonce2.owner = shared_channel_R_to_A[0].nonce2.owner;
		      msg[0].nonce2.value = shared_channel_R_to_A[0].nonce2.value;
		      msg[0].address = shared_channel_R_to_A[0].address;
		      shared_channel_R_to_A_filled = 0;
	      } else {
		      wait_process();
	      }
	      */
	      while (shared_channel_R_to_A_filled != 1) { 
		wait_process();
	      }
              {
		      msg[0].key = shared_channel_R_to_A[0].key;
		      msg[0].nonce1.owner = shared_channel_R_to_A[0].nonce1.owner;
		      msg[0].nonce1.value = shared_channel_R_to_A[0].nonce1.value;
		      msg[0].nonce2.owner = shared_channel_R_to_A[0].nonce2.owner;
		      msg[0].nonce2.value = shared_channel_R_to_A[0].nonce2.value;
		      msg[0].address = shared_channel_R_to_A[0].address;
		      shared_channel_R_to_A_filled = 0;
	      }
              successful = 1;
	  }
          if (successful) {
              messages[nextspace].key = msg[0].key;
              messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
              messages[nextspace].nonce1.value = msg[0].nonce1.value;
              messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
              messages[nextspace].nonce2.value = msg[0].nonce2.value;
	      messages[nextspace].address = msg[0].address;
              ++nextspace;
              if (msg[0].key == A) {
              	/* learn nonces; no check for indices in range */
              	nonces[msg[0].nonce1.owner][msg[0].nonce1.value] = 1;
             	nonces[msg[0].nonce2.owner][msg[0].nonce2.value] = 1;             
	      }
	  }
	}
     } else if (rand == 1) {
    /* attempt to send msg */
        int  target = VS_toss(MAXKNOWLEDGE - 1);

        if (target < nextspace) {
	      msg[0].key = messages[target].key;
	      msg[0].nonce1.owner = messages[target].nonce1.owner;
	      msg[0].nonce1.value = messages[target].nonce1.value;
	      msg[0].nonce2.owner = messages[target].nonce2.owner;
	      msg[0].nonce2.value = messages[target].nonce2.value;
	      msg[0].address = messages[target].address;
          /* smsg is malloc'ed in MsgToSmsg, so free it below */
          if (VS_toss(1) == 0) {
	     /*
	     if (shared_channel_A_to_I_filled == 0) {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_A_to_I (a1, a2, a3, a4, a5, a6);
		shared_channel_A_to_I_filled = 1;
		//wait_process();
	   } else {
		wait_process();
	   }
	   */
	   while (shared_channel_A_to_I_filled != 0) {
		wait_process ();
	   }
	   {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_A_to_I (a1, a2, a3, a4, a5, a6);
		shared_channel_A_to_I_filled = 1;
		wait_process();
	   }
	} else {
	   /*
	   if (shared_channel_A_to_R_filled == 0) {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_A_to_R (a1, a2, a3, a4, a5, a6);
		shared_channel_A_to_R_filled = 1;
		//wait_process();
	   } else {
		wait_process();
	   }
	   */
	   while (shared_channel_A_to_R_filled != 0) {
		wait_process ();
	   }
	   {
		AgentId a1, a2, a4, a6;
		int a3, a5;

		a1 = msg[0].key;
		a2 = msg[0].nonce1.owner;
		a3 = msg[0].nonce1.value;
		a4 = msg[0].nonce2.owner;
		a5 = msg[0].nonce2.value;
		a6 = msg[0].address;
		send_shared_channel_A_to_R (a1, a2, a3, a4, a5, a6);
		shared_channel_A_to_R_filled = 1;
		wait_process();
	   }
	}
      }
    } else if (rand == 2) {
    /* create or dump a message */
        if (nextspace >= MAXKNOWLEDGE && DUMP) {  
          /* dump a message */
          messages[VS_toss(MAXKNOWLEDGE - 2)].key = messages[MAXKNOWLEDGE - 1].key;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce1.owner = messages[MAXKNOWLEDGE - 1].nonce1.owner;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce1.value = messages[MAXKNOWLEDGE - 1].nonce1.value;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce2.owner = messages[MAXKNOWLEDGE - 1].nonce2.owner;
          messages[VS_toss(MAXKNOWLEDGE - 2)].nonce2.value = messages[MAXKNOWLEDGE - 1].nonce2.value;
          messages[VS_toss(MAXKNOWLEDGE - 2)].address = messages[MAXKNOWLEDGE - 1].address;
          nextspace = MAXKNOWLEDGE - 1;
	}
        else if (nextspace < MAXKNOWLEDGE) {
          /* attempt to create a message using known nonces */
          int     duplicate = 0, i;
          AgentId k, n1o, n2o, a;
          int     n1v, n2v;

          k = VS_toss(NUMAGENTS - 3) + 1; 
          /* The following logic ensures that the adversary generates 
             only valid message forms.  Forms that will be rejected 
             by the honest agents as bogus are avoided in order to prune
             the state space.  By nesting the logic, the number of 
             calls to VS_toss is reduced.  */
          if (k == I) {
            /* form must be N1,N2,E, where N1 is a defined nonce 
               owned by I, N2 is defined, E is undefined if 
               !FIXED, and E != I if FIXED */
            n1o = I;
            n1v = VS_toss(NUMNONCES - 1) + 1; /* n1v in 1..NUMNONCES */
            if (VS_toss(1) == 0) {
            	n2o = R;
	    } else {
            	n2o = A;
	    }
            n2v = (VS_toss(NUMNONCES - 1)) + 1;
	    if (!FIXED) {
              a = D;
	    }
            else {
              if (VS_toss(1) == 0) {
              	a = R;
	      } else {
              	a = A;
	      }
	    }
	    if (nonces[n1o][n1v] && nonces[n2o][n2v]) {  /* nonces are known */
		    msg[0].key = k;
		    msg[0].nonce1.owner = n1o;
		    msg[0].nonce1.value = n1v;
		    msg[0].nonce2.owner = n2o;
		    msg[0].nonce2.value = n2v;
		    msg[0].address = a;
		    for (i = 0; i < nextspace; ++i) {
		      	duplicate = duplicate || (msg[0].key == messages[i].key 
						  && msg[0].nonce1.owner == messages[i].nonce1.owner
						  && msg[0].nonce1.value == messages[i].nonce1.value
						  && msg[0].nonce2.owner == messages[i].nonce2.owner
						  && msg[0].nonce1.value == messages[i].nonce2.value);
		
		    }
		    if (!duplicate) {
		    	messages[nextspace].key = msg[0].key;
		    	messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
		    	messages[nextspace].nonce1.value = msg[0].nonce1.value;
		    	messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
		    	messages[nextspace].nonce2.value = msg[0].nonce2.value;
		    	messages[nextspace].address = msg[0].address;
		    	++nextspace;
		    }
	    }
	  }
          else if (k == R) {
            /* form must be N1,U,E, where U is undefined, N1 is a 
               defined nonce, E undefined implies N1 is owned by R,
               and E defined implies E != R */
            n2o = D;
            n2v = 0;

            if (VS_toss(1) == 0) {
            /* E undefined */
              a = D;
              n1o = R;
              n1v = VS_toss(NUMNONCES - 1) + 1;
	    } else {
            /* E defined */
              if (VS_toss(1) == 0) {
              	a = I;
	       } else {
              	a = A;
	       }
              if (VS_toss(1) == 0) {
              	n1o = I;
	      }else {
              	n1o = A;
              }
              n1v = VS_toss(NUMNONCES - 1) + 1;                  
            }
	    if (nonces[n1o][n1v] && nonces[n2o][n2v])  {  /* nonces are known */
	      msg[0].key = k;
	      msg[0].nonce1.owner = n1o;
	      msg[0].nonce1.value = n1v;
	      msg[0].nonce2.owner = n2o;
	      msg[0].nonce2.value = n2v;
	      msg[0].address = a;
	      for (i = 0; i < nextspace; ++i) {
	      		duplicate = duplicate || (msg[0].key == messages[i].key 
					  && msg[0].nonce1.owner == messages[i].nonce1.owner
					  && msg[0].nonce1.value == messages[i].nonce1.value
					  && msg[0].nonce2.owner == messages[i].nonce2.owner
					  && msg[0].nonce1.value == messages[i].nonce2.value);
	      }
	      if (!duplicate) {
		    messages[nextspace].key = msg[0].key;
		    messages[nextspace].nonce1.owner = msg[0].nonce1.owner;
		    messages[nextspace].nonce1.value = msg[0].nonce1.value;
		    messages[nextspace].nonce2.owner = msg[0].nonce2.owner;
		    messages[nextspace].nonce2.value = msg[0].nonce2.value;
		    messages[nextspace].address = msg[0].address;
		    ++nextspace;
	       }
	     }
          }
        }
    }
  }
}

int VS_toss (int x) {

}

void send_shared_channel_I_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
	shared_channel_I_to_A[0].key = a1;
	shared_channel_I_to_A[0].nonce1.owner = a2;
	shared_channel_I_to_A[0].nonce1.value = a3;
	shared_channel_I_to_A[0].nonce2.owner = a4;
	shared_channel_I_to_A[0].nonce2.value = a5;
	shared_channel_I_to_A[0].address = a6;
}

void send_shared_channel_A_to_I (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
	shared_channel_A_to_I[0].key = a1;
	shared_channel_A_to_I[0].nonce1.owner = a2;
	shared_channel_A_to_I[0].nonce1.value = a3;
	shared_channel_A_to_I[0].nonce2.owner = a4;
	shared_channel_A_to_I[0].nonce2.value = a5;
	shared_channel_A_to_I[0].address = a6;
}

void send_shared_channel_R_to_A (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
	shared_channel_R_to_A[0].key = a1;
	shared_channel_R_to_A[0].nonce1.owner = a2;
	shared_channel_R_to_A[0].nonce1.value = a3;
	shared_channel_R_to_A[0].nonce2.owner = a4;
	shared_channel_R_to_A[0].nonce2.value = a5;
	shared_channel_R_to_A[0].address = a6;
}

void send_shared_channel_A_to_R (AgentId a1, AgentId a2, int a3, AgentId a4, int a5, AgentId a6) {
	shared_channel_A_to_R[0].key = a1;
	shared_channel_A_to_R[0].nonce1.owner = a2;
	shared_channel_A_to_R[0].nonce1.value = a3;
	shared_channel_A_to_R[0].nonce2.owner = a4;
	shared_channel_A_to_R[0].nonce2.value = a5;
	shared_channel_A_to_R[0].address = a6;
}

void wait_process() {

}

int VS_assert (int exp) {
	if (exp == 0)
		return 0;
	else 
		return 1;
}
